local mod = LibStub("AceAddon-3.0"):NewAddon("BigBrother_Ready", "AceEvent-3.0", "AceTimer-3.0")

--[===[@debug@
_G.BBReady = mod

local PREFIX = "|cff00ff00BB_Ready|r: "
local DevTools_Dump
DevTools_Dump = function(value)
	if not _G.DevTools_Dump then
		LoadAddOn("Blizzard_DebugTools")
	end
	DevTools_Dump = _G.DevTools_Dump
	DevTools_Dump(value)
end

local function dump(value)
	local orig = DEFAULT_CHAT_FRAME
	DEFAULT_CHAT_FRAME = ChatFrame4
	DevTools_Dump(value)
	DEFAULT_CHAT_FRAME = orig
end

local function mystrjoin(arg1, ...)
	local text = tostring(arg1)
	for i = 1, select('#', ...) do
		text = text .. tostring(select(i, ...))
	end
	return text
end

local doDebug
local function realDebug(arg1, ...)
	if not arg1 then return end
	if type(arg1) == 'table' then
		ChatFrame4:AddMessage(PREFIX .. "+++ table-dump")
		dump(arg1)
		ChatFrame4:AddMessage(PREFIX .. "--- end of table-dump")
		realDebug(...)
	else
		local text = mystrjoin(PREFIX, arg1, ...)
		ChatFrame4:AddMessage(text)
	end
end

local buffer = {}
local function collectDebug(...)
	buffer[#buffer + 1] = { ... }
end
doDebug = collectDebug

local function releaseDebug(self)
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	doDebug = realDebug
	for k,v in ipairs(buffer) do
		doDebug(unpack(v))
	end
	wipe(buffer)
	buffer = nil
end
local DebugFrame = CreateFrame('Frame')
DebugFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
DebugFrame:SetScript("OnEvent", releaseDebug)
--@end-debug@]===]

local function debug(...)
	--[===[@debug@
	doDebug(...)
	--@end-debug@]===]
end
mod.debug = debug

local foods = {
	35272, -- Well Fed
	44106, -- "Well Fed" from Brewfest
}

local flasks = {
	-- Flask WotLK
	53752, -- 53752 Lesser Flask of Toughness
	53755, -- 53755 Flask of the Frost Wyrm
	53758, -- 53758 Flask of Stoneblood
	53760, -- 53760 Flask of Pure Mojo
	54212, -- 54212 Flask of Endless Rage
	62380, -- Lesser Flask of Resistance
--	67016, -- Flask of North
}

local guardianElixirs = {
	53747, -- 53747 Elixir of Spirit
	60347, -- 60347 Elixir of Mighty Thoughts
	53764, -- 53764 Elixir of Mighty Mageblood
	53751, -- 53751 Elixir of Mighty Fortitude
	60343, -- 60343 Elixir of Mighty Defense
}

local battleElixirs = {
	53746, -- 53746 Wrath Elixir
	53749, -- 53749 Guru's Elixir
	53763, -- 53763 Elixir of Protection
	53748, -- 53748 Elixir of Mighty Strength
	28497, -- 53748 Elixir of Mighty Agility
	60346, -- 60346 Elixir of Lightning Speed
	60344, -- 60344 Elixir of Expertise
	60341, -- 60341 Elixir of Deadly Strikes
	60345, -- 60345 Elixir of Armor Piercing
	60340, -- 60340 Elixir of Accuracy
}

local db

local MIN_DURATION = 10*60
local function checkForAnyBuff(unit, buffList)
	for i,v in ipairs(buffList) do
		local name, _, _, _, _, duration, expirationTime = UnitAura(unit, GetSpellInfo(v) or v)
		if name then
			if expirationTime then
				local remaining = expirationTime - GetTime()
				return remaining
			else
				return
			end
		end
	end
	return
end

local function checkForMissingElixir(unit)
	local guardianRemaining = checkForAnyBuff(unit, guardianElixirs)
	local battleRemaining = checkForAnyBuff(unit, battleElixirs)
	if guardianRemaining and battleRemaining then
		if guardianRemaining > MIN_DURATION and battleRemaining > MIN_DURATION then
			return
		elseif guardianRemaining < MIN_DURATION or battleRemaining < MIN_DURATION then
			return "Expiring elixirs"
		end
	end
	local onlybuff = guardianRemaining or battleRemaining
	if onlybuff then
		if onlybuff < MIN_DURATION then
			return "Only one expiring elixir"
		else
			return "Only one elixir"
		end
	else
		return "No flask or elixir"
	end
	return false
end

local function checkForMissingFlask(unit)
	local remaining = checkForAnyBuff(unit, flasks)
	if not remaining then
		return checkForMissingElixir(unit)
	elseif remaining < MIN_DURATION then
		return "Expiring Flask"
	else
		return
	end
end

local function checkForMissingFood(unit)
	local remaining = checkForAnyBuff(unit, foods)
	if remaining then
		if remaining < MIN_DURATION then
			return "Expiring food buff"
		else
			return
		end
	else
		return "No food buff"
	end
end


local function broadCastToRaid(prefix, table)
	if #table > 0 then
		SendChatMessage(prefix .. ": " .. strjoin(", ", unpack(table)), "RAID")
	end
end
local TOKEN = "<BB_Ready>"
local outgoingFilter = function(self, event, msg, ...)
	if msg:find(TOKEN) then return true end
end
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", outgoingFilter)

local function broadCastWhisper(state, table)
	for _, v in ipairs(table) do
		SendChatMessage(("%s %s"):format(TOKEN, state), "WHISPER", nil, v)
	end
end

local function autoInitWithEmptyTable(self, key)
	local val = rawget(self, key)
	if not val then
		val = {}
		rawset(self, key, val)
	end
	return val
end
local mt = { __index = autoInitWithEmptyTable, }

function mod:READY_CHECK()
	local buffStates = setmetatable({}, mt)
	for i = 1, GetNumRaidMembers() do
		local unit = "raid" .. i
		local name = UnitName(unit)
		if not UnitIsVisible(unit) then
			table.insert(buffStates["out of range"], name)
		else
			local state = checkForMissingFlask(unit)
			if state then
				table.insert(buffStates[state], name)
			end
			local foodstate = checkForMissingFood(unit)
			if foodstate then
				table.insert(buffStates[foodstate], name)
			end
		end
	end
	debug("buffStates")
	debug(buffStates)
	for k,v in pairs(buffStates) do
		broadCastToRaid(k, v)
	end
	if db.state == 2 then
		wipe(buffStates["out of range"])
		for k,v in pairs(buffStates) do
			broadCastWhisper(k, v)
		end
	end
	-- cleanup the mess
	wipe(buffStates)
	buffStates = nil
end

local function checkForBoss()
	-- TODO implement
	return false
end

function mod:PLAYER_REGEN_ENABLED()
	if checkForBoss() then
		self:ScheduleTimer("READY_CHECK", 10)
	end
end

local dbDefaults = {
	profile = {
		state = 0,
	}
}

function mod:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("BigBrother_ReadyDB", dbDefaults)
	db = self.db.profile
	self:SetIcon()
end

function mod:OnEnable()
	debug("enabling")
	if db.state == 0 then
		return
	end
	self:SetupEvents()
end

function mod:SetupEvents()
	self:RegisterEvent("CHAT_MSG_ADDON")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "UpdateMembers")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "UpdateMembers")
end

function mod:Resume()
	self:SetupEvents()
end

function mod:Suspend()
	self:UnregisterAllEvents()
	self:CancelAllTimers()
end

function mod:OnDisable()
	self:Suspend()
	self:Resync()
end

local stateIcons = {
	[0] = [[Interface\RAIDFRAME\ReadyCheck-NotReady]],
	[1] = [[Interface\RAIDFRAME\ReadyCheck-Ready]],
	[2] = [[Interface\RAIDFRAME\ReadyCheck-Ready]],
	-- [2] = TODO aggressive icon
}

local stateTexts = {
	[0] = "Off",
	[1] = "On",
	[2] = "On*",
}

local ldbdata
function mod:SetIcon()
	ldbdata.icon = stateIcons[db.state]
	ldbdata.text = stateTexts[db.state]
end

function mod:SetState(state)
	mod:SetIcon()
	if state == 0 then
		mod:Suspend()
	elseif state == 1 then
		mod:Resume()
	end
	mod:Resync()
end

local function incrementState()
	db.state = (db.state + 1) % (#stateIcons + 1)
	debug("state is now ", db.state)
	mod:SetState(db.state)
end

ldbdata = {
	type = "launcher",
	OnClick = function(_, msg)
			if msg == "LeftButton" then
				incrementState()
			elseif msg == "RightButton" then
				mod:READY_CHECK()
			end
		end,
	label = "BigBrother_Ready",
}

LibStub("LibDataBroker-1.1"):NewDataObject("BigBrother_Ready", ldbdata)

-- sync-stuff
local candidates = {}
local pedantic = false
local ADDON_MSG_PREFIX = "BBReady"
local playerName = UnitName("player")

function mod:Resync()
	debug("resyncing")
	SendAddonMessage(ADDON_MSG_PREFIX, "resync", "RAID")
end

local timerHandle
function mod:ChooseCandidate()
	debug("timer expired")
	if #candidates > 0 then
		table.sort(candidates)
		debug(candidates)
		--debug(candidates[1], " ~= ", play
		if playerName == candidates[1] then
			debug("yesss, it's me")
			self:RegisterEvent("READY_CHECK")
		else
			debug(":( it's not me")
			self:UnregisterEvent("READY_CHECK")
		end
	else
		debug("yesss, it's me")
		self:RegisterEvent("READY_CHECK")
	end
	debug(candidates[1], " was chosen")
	wipe(candidates)
	pedantic = false
	timerHandle = nil
end

local pingTimer
local function sendPing()
	debug("ping the raid")
	local message = db.state == 2 and "pingping" or "ping"
	SendAddonMessage(ADDON_MSG_PREFIX, message, "RAID")
	pingTimer = nil
end

local raidMembers = {}
local function doUpdateMembers()
	local shouldUpdate = false
	local missingMembers = CopyTable(raidMembers)
	for i = 1,GetNumRaidMembers() do
		local unit = "raid" .. i
		local name = UnitName(unit)
		if not raidMembers[name] then
			raidMembers[name] = true
		end
		missingMembers[name] = nil
	end
	for k in pairs(missingMembers) do
		raidMembers[k] = nil
		missingMembers[k] = nil
		shouldUpdate = true
	end
	return shouldUpdate
end
function mod:UpdateMembers()
	debug("udpating members ", GetNumRaidMembers())
	if not next(raidMembers) then
		doUpdateMembers()
		self:Resync()
	elseif doUpdateMembers() then
		self:CheckGroup()
	end
end

function mod:CheckGroup(arg, ...)
	pedantic = false
	if timerHandle then
		debug("reschedule")
		wipe(candidates)
		self:CancelTimer(timerHandle)
	end
	debug("schedule timer")
	timerHandle = self:ScheduleTimer("ChooseCandidate", 10)
	if pingTimer then
		debug("reschedule ping-timer")
		self:CancelTimer(pingTimer)
	end
	pingTimer = self:ScheduleTimer(sendPing, 2)
end
function mod:CHAT_MSG_ADDON(event, prefix, message, _, sender)
	--debug(prefix, " ", message, " ", sender)
	if prefix == ADDON_MSG_PREFIX then
		debug("message: ", message, " (", sender, ")")
		if message == "pingping" then
			if not pedantic then
				wipe(candidates)
				pedantic = true
			end
			table.insert(candidates, sender)
		elseif message == "ping" and not pedantic then
			table.insert(candidates, sender)
		elseif message == "resync" then
			debug("recieved resync")
			self:CheckGroup()
		end
	end
end
